/* 
Expression parser / evaluator.
Evaluate f(v0, v1, v2, v3, ... ) for provided v[].  
F = Expression of + - * / ^ ( ) , sgn abs sin cos ln exp max min ... 
'-' preceded by operator or '(' is negate, otherwise it is minus
Ex : 
	CExpression o;
	double a[1];

	o.Parse("t^3-60*t^2+1100*t-6000+abs(sin(t/5)*sin(t/4)*100+90*cos(t/6)*cos(t/3))", "t");

	a[0]=0.1234;
	double t1=o.Evaluate(a);
	a[0]=9876.54;
	double t2=o.Evaluate(a);
	
	todo - change all token chars to BYTE's	
	todo - consider out of order var match in parse if any varnames are prefix of others.
			
Expression examples  : 
	6.5 + 3
	6.5 - 3
	6.5 * 3
	6.5 / 3
	6.5 ^ 3
	max( 6.5, v ) 
	min( 6.5, v ) 
	sin( v ) 
	cos( v )	
	ln( v )			natural log
	exp( v )		e raised to v power
	sgn( v )		Returns -1, 0, 1 for values <0, =0, >0 
	rnd( u )		Returns random value [0..u)
	abs( v )		absolute value
	line(u,v)		line from u to v 
	ramp(0.7,u,v,w)	line from u to v during initial 7/10, from v to w during final 3/10
	rmpex( a, v1, v2 )	exponential e^(at) rise from v1 to v2, fall from v2 to v1
	pulse(0.7,u,v)	value u during first 7/10 of segment, value v during final 3/10 of seg
	square( .1, .2, v1, v2 )	v1 for first .100 and last .200 ; v2 between.
	any combination of the above, i.e. : 
		5 * sgn( v + line(u,v) + 4^u - u/( sin(t*3) + i ) )
	hai( r, v1,t1,v2,t2,v3 ) - Series of pulses of 3 different amplitudes.
		Pulse amplitude dependant on iteration For t=0..t1, amplitude v1; 
		t={t1..t2} v2; t={t2..1.0} v3;
		ramp up to and down from pulse during r portion of t( i.e r=0.1 - ramp up and down 
		during first and last 10 percent )
*/

class CExpression { 
	private : 
		int m_iTokens, m_iStackSize, m_iVars;
		int * m_aTokens;		// Expression parsed and reordered w/o parentheses
		int * m_aRaw;			// Expression parsed in original order 
		double * m_aValues;
		double * m_aStack;
		char * m_achVarnames;
		char ** m_apchVars;	
		bool IsFunction( int iToken );
		bool IsValue( int iToken );
		bool IsBinary( int iToken );
		int ParseFind( int iToken1, int iToken2, int i0, int i1 ); 
		int ParseFindL( int iToken1, int iToken2, int i0, int i1 );
		void ParseShiftL( int i0, int i1, int iShift=1 );
		void ParseReduce( int i0, int i1 );
		bool NextToken( const char * pchExpr, int * piPos, int * piToken, int iPrevToken, double * pnValue );

	public :
		bool m_bValid;
		CExpression( const char * pchExpr, char * achVariables  );	// note achVariables not a const, but is not altered
		CExpression();
		~CExpression();

		int FindVar( int iLow, int iHigh );
					
		bool Parse( const char * pchExpr, char * achVariables  ); 
		bool ParseVariables( char * pch );
		double Evaluate( double * aVar, bool * pbError=NULL );

		#ifdef _DEBUG
		//----- This function displays the parse results for debug purposes
		void Text( char * achBuffer, int iBufSize );
		#endif
	};
	